/*
 * Copyright (C) 1998-2001 Patric Mller
 * bhaak@gmx.net
 * http://www.mysunrise.ch/users/bhaak/vilistextum/
 *
 * Released under the GNU GPL - http://www.gnu.org/copyleft/gpl.html
 *
 */

/*  History:
 *  18.04.01: now ignores entities for ascii control characters (0-31)
 *  19.04.01: added parse_entities(char *)
 *  03.09.01: added hexadecimal entities
 *
 */

#include <../config.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>      

#include "latin1.h"
#include "text.h"
#include "debug.h"
#include "util.h"
#include "main.h"
#include "microsoft.h"
#include "charset.h"

#include "multibyte.h"

int entity_number(CHAR *s);
int html_entity(CHAR*);
int latin1(CHAR*);

// ------------------------------------------------

void set_char_wrapper(CHAR *str, int num)
{
#ifdef MULTIBYTE
	printf("Multibyte\n");
	convert_character(num, str);
#else	
# if DEBUG
	printf("No Multibyte\n");
# endif
	set_char(str, num);
#endif
} // end set_char_wrapper

// ------------------------------------------------

// parse entity in string 
void parse_entity(CHAR *str)
{
	int i;
	CHAR tmpstr[DEF_STR_LEN];
	int len = STRLEN(str);
#if DEBUG
	for (i=0; i<10; i++) { printf("%d ", str[i]);}
#endif
#ifdef proc_debug
	printf("start parse: %s\n", str);
#endif	

	for (i=0; i<DEF_STR_LEN; i++) { tmpstr[i]=0x00; }
	CPYSS(tmpstr, str);

#if DEBUG
	printf("1 %s\n", str);
	printf("2 %s\n", tmpstr);

	for (i=0; i<10; i++) { printf("%d ", *(str+i));}
	for (i=0; i<10; i++) { printf("%d ", *(tmpstr+i));}
	printf("\n");
#endif
	//	fprintf(stderr, "tmpstr: %s-\n", tmpstr);
	//fprintf(stderr, "tmpstr: %d-\n", tmpstr[len]);
	if (tmpstr[len-1]!=';') 
  {
		tmpstr[len]   = ';';
		tmpstr[len+1] = '\0';
	}
	//	printf("tmpstr: %s\n", tmpstr);

	if ((entity_number(tmpstr))||(html_entity(tmpstr))||(latin1(tmpstr))||(microsoft_entities(tmpstr))) 
	{
		// if true entity was known
		CPYSS(str, tmpstr);
		//fprintf(stderr, "Entity String %s\n", str);
	}
	else
	{
		if (errorlevel>=1) { print_error("entity unknown: ", tmpstr); }
	}
	
	//	fprintf(stderr, "Entity String %s %s\n", str, tmpstr);
#ifdef proc_debug
	printf("end parse: %s\n", str);
#endif
} // end parse

// ------------------------------------------------

// parses entities in string
void parse_entities(CHAR *s)
{
	int i=0,j=0,k=0;
	CHAR tmp[DEF_STR_LEN];
	CHAR entity[DEF_STR_LEN];
	int len=STRLEN(s);
	CHAR result[DEF_STR_LEN];
	
	result[0] = '\0';

	//printf("Ganzer String -%s-\n", s);
	
	while(i<=len)
	{
		j=0;
		while((s[i]!='\0') && (s[i]!='&'))
	  {
			tmp[j++] = s[i++];
		}
		tmp[j] = '\0';
		STRCAT(result, tmp);

		if (s[i]=='&')
		{
		  k=0;
			//entity[0] = '&';
			while((s[i]!='\0') && (s[i]!=';') && (!isspace(s[i])))
	    {
				entity[k++] = s[i++];
			}
			entity[k] = '\0';
		parse_entity(entity);

			STRCAT(result, entity);
		}
		i++;
	}

	//printf("Result: %s\n", result);

	CPYSS(s, result);
} // end parse_entities

// ------------------------------------------------

int entity_number(CHAR *s)
{
  int number;
  CHAR *tmp = s;

  // Numeric entity
  if ((s[0]=='&') && (s[1]=='#'))
	{	   
		// Hex entity
		if (uppercase(s[2])=='X')
		{
			tmp += 3;
			number = x2dec(tmp, 16);					 
		} 
		// Decimal entity
		else {
			tmp += 2;
			number = ATOI(tmp);	
		}

#ifdef MULTIBYTE
		convert_character(number, s);
		return(1);
#else
		// ascii printable character 32-127 
		if ((number>=32) && (number<=127)) 
		{ 
			set_char(s, number);
			return(1);
		}
		// ansi printable character 160-255
		else if ((number>=160) && (number<=255)) 
		{
			//fprintf(stderr, "1 sevenbit %d\n", sevenbit);
			if (sevenbit) { return(0); }
			set_char(s, number);
			return(1);
		}
		// ascii control character -> return empty string
		else if ((number >= 0) && (number < 32)) 
		{ 
			s[0] = '\0';
			return(1); 
		}
#endif
	}
  return(0);
} // end entity_number

// ------------------------------------------------

int html_entity(CHAR *str)
{	
  if (sevenbit)
  {
	// clutch until proper charset conversion is done
	if CMP("&nbsp;", str)      { set_char(str, ' '); } // no-break space 
	else if CMP("&#160;", str) { set_char(str, ' '); } // no-break space 
  }

  if CMP("&quot;",     str)    { set_char(str, '"'); }
  else if CMP("&;",    str)    { set_char(str, '&'); } // for those brain damaged ones
  else if CMP("&amp;", str)    { set_char(str, '&'); }
  else if CMP("&gt;",  str)    { set_char(str, '>'); }
  else if CMP("&lt;",  str)    { set_char(str, '<'); }
  else { return(0); } // found no html entity
  return(1); // found html entity 
} // end html_entity

// ------------------------------------------------

// returns true if str is a known entity and changes str to the printout
int latin1(CHAR *str)
{
  //fprintf(stderr, "2 sevenbit %d\n", sevenbit);
  //if (sevenbit) { return(0); } // leave char>127 alone

  if CMP("&nbsp;", str)         { set_char(str, 160); } // no-break space 
  else if CMP("&iexcl", str)    { set_char(str, 161); } // inverted exclamation mark 
  else if CMP("&cent;", str)    { set_char(str, 162); } // cent sign 
  else if CMP("&pound;", str)   { set_char(str, 163); } // pound sterling sign 
  else if CMP("&curren;", str)  { set_char(str, 164); } // general currency sign 
  else if CMP("&yen;", str)     { set_char(str, 165); } // yen sign 
  else if CMP("&brvbar;", str)  { set_char(str, 166); } // broken (vertical) bar 
  else if CMP("&sect;", str)    { set_char(str, 167); } // section sign 
  else if CMP("&uml;", str)     { set_char(str, 168); } // umlaut (dieresis) 
  else if CMP("&copy;", str)    { set_char(str, 169); } // copyright sign 
  else if CMP("&ordf;", str)    { set_char(str, 170); } // ordinal indicator, feminine 
  else if CMP("&laquo;", str)   { set_char(str, 171); } // angle quotation mark, left 
  else if CMP("&not;", str)     { set_char(str, 172); } // not sign 
  else if CMP("&shy;", str)     { set_char(str, 173); } // soft hyphen 
  else if CMP("&reg;", str)     { set_char(str, 174); } // registered sign 
  else if CMP("&macr;", str)    { set_char(str, 175); } // macron 
  else if CMP("&deg;", str)     { set_char(str, 176); } // degree sign 
  else if CMP("&plusmn;", str)  { set_char(str, 177); } // plus-or-minus sign 
  else if CMP("&sup2;", str)    { set_char(str, 178); } // superscript two 
  else if CMP("&sup3;", str)    { set_char(str, 179); } // superscript three 
  else if CMP("&acute;", str)   { set_char(str, 180); } // acute accent 
  else if CMP("&micro;", str)   { set_char(str, 181); } // micro sign 
  else if CMP("&para;", str)    { set_char(str, 182); } // pilcrow (paragraph sign) 
  else if CMP("&middot;", str)  { set_char(str, 183); } // middle dot 
  else if CMP("&cedil;", str)   { set_char(str, 184); } // cedilla 
  else if CMP("&sup1;", str)    { set_char(str, 185); } // superscript one 
  else if CMP("&ordm;", str)    { set_char(str, 186); } // ordinal indicator, masculine 
  else if CMP("&raquo;", str)   { set_char(str, 187); } // angle quotation mark, right 
  else if CMP("&frac14;", str)  { set_char(str, 188); } // fraction one-quarter 
  else if CMP("&frac12;", str)  { set_char(str, 189); } // fraction one-half 
  else if CMP("&frac34;", str)  { set_char(str, 190); } // fraction three-quarters 
  else if CMP("&iquest;", str)  { set_char(str, 191); } // inverted question mark 
  else if CMP("&Agrave;", str)  { set_char(str, 192); } // capital A, grave accent 
  else if CMP("&Aacute;", str)  { set_char(str, 193); } // capital A, acute accent 
  else if CMP("&Acirc;", str)   { set_char(str, 194); } // capital A, circumflex accent 
  else if CMP("&Atilde;", str)  { set_char(str, 195); } // capital A, tilde 
  else if CMP("&Auml;", str)    { set_char(str, 196); } // capital A, dieresis or umlaut mark 
  else if CMP("&Aring;", str)   { set_char(str, 197); } // capital A, ring 
  else if CMP("&AElig;", str)   { set_char(str, 198); } // capital AE diphthong (ligature) 
  else if CMP("&Ccedil;", str)  { set_char(str, 199); } // capital C, cedilla 
  else if CMP("&Egrave;", str)  { set_char(str, 200); } // capital E, grave accent 
  else if CMP("&Eacute;", str)  { set_char(str, 201); } // capital E, acute accent 
  else if CMP("&Ecirc;", str)   { set_char(str, 202); } // capital E, circumflex accent 
  else if CMP("&Euml;", str)    { set_char(str, 203); } // capital E, dieresis or umlaut mark 
  else if CMP("&Igrave;", str)  { set_char(str, 204); } // capital I, grave accent 
  else if CMP("&Iacute;", str)  { set_char(str, 205); } // capital I, acute accent 
  else if CMP("&Icirc;", str)   { set_char(str, 206); } // capital I, circumflex accent 
  else if CMP("&Iuml;", str)    { set_char(str, 207); } // capital I, dieresis or umlaut mark 
  else if CMP("&ETH;", str)     { set_char(str, 208); } // capital Eth, Icelandic 
  else if CMP("&Ntilde;", str)  { set_char(str, 209); } // capital N, tilde 
  else if CMP("&Ograve;", str)  { set_char(str, 210); } // capital O, grave accent 
  else if CMP("&Oacute;", str)  { set_char(str, 211); } // capital O, acute accent 
  else if CMP("&Ocirc;", str)   { set_char(str, 212); } // capital O, circumflex accent 
  else if CMP("&Otilde;", str)  { set_char(str, 213); } // capital O, tilde 
  else if CMP("&Ouml;", str)    { set_char(str, 214); } // capital O, dieresis or umlaut mark 
  else if CMP("&times;", str)   { set_char(str, 215); } // multiply sign 
  else if CMP("&Oslash;", str)  { set_char(str, 216); } // capital O, slash 
  else if CMP("&Ugrave;", str)  { set_char(str, 217); } // capital U, grave accent 
  else if CMP("&Uacute;", str)  { set_char(str, 218); } // capital U, acute accent 
  else if CMP("&Ucirc;", str)   { set_char(str, 219); } // capital U, circumflex accent 
  else if CMP("&Uuml;", str)    { set_char(str, 220); } // capital U, dieresis or umlaut mark 
  else if CMP("&Yacute;", str)  { set_char(str, 221); } // capital Y, acute accent 
  else if CMP("&THORN;", str)   { set_char(str, 222); } // capital THORN, Icelandic 
  else if CMP("&szlig;", str)   { set_char(str, 223); } // small sharp s, German (sz ligature) 
  else if CMP("&agrave;", str)  { set_char(str, 224); } // small a, grave accent 
  else if CMP("&aacute;", str)  { set_char(str, 225); } // small a, acute accent 
  else if CMP("&acirc;", str)   { set_char(str, 226); } // small a, circumflex accent 
  else if CMP("&atilde;", str)  { set_char(str, 227); } // small a, tilde 
  else if CMP("&auml;", str)    { set_char(str, 228); } // small a, dieresis or umlaut mark 
  else if CMP("&aring;", str)   { set_char(str, 229); } // small a, ring 
  else if CMP("&aelig;", str)   { set_char(str, 230); } // small ae diphthong (ligature) 
  else if CMP("&ccedil;", str)  { set_char(str, 231); } // small c, cedilla 
  else if CMP("&egrave;", str)  { set_char(str, 232); } // small e, grave accent 
  else if CMP("&eacute;", str)  { set_char(str, 233); } // small e, acute accent 
  else if CMP("&ecirc;", str)   { set_char(str, 234); } // small e, circumflex accent 
  else if CMP("&euml;", str)    { set_char(str, 235); } // small e, dieresis or umlaut mark 
  else if CMP("&igrave;", str)  { set_char(str, 236); } // small i, grave accent 
  else if CMP("&iacute;", str)  { set_char(str, 237); } // small i, acute accent 
  else if CMP("&icirc;", str)   { set_char(str, 238); } // small i, circumflex accent 
  else if CMP("&iuml;", str)    { set_char(str, 239); } // small i, dieresis or umlaut mark 
  else if CMP("&eth;", str)     { set_char(str, 240); } // small eth, Icelandic 
  else if CMP("&ntilde;", str)  { set_char(str, 241); } // small n, tilde 
  else if CMP("&ograve;", str)  { set_char(str, 242); } // small o, grave accent 
  else if CMP("&oacute;", str)  { set_char(str, 243); } // small o, acute accent 
  else if CMP("&ocirc;", str)   { set_char(str, 244); } // small o, circumflex accent 
  else if CMP("&otilde;", str)  { set_char(str, 245); } // small o, tilde 
  else if CMP("&ouml;", str)    { set_char(str, 246); } // small o, dieresis or umlaut mark 
  else if CMP("&divide;", str)  { set_char(str, 247); } // divide sign 
  else if CMP("&oslash;", str)  { set_char(str, 248); } // small o, slash 
  else if CMP("&ugrave;", str)  { set_char(str, 249); } // small u, grave accent 
  else if CMP("&uacute;", str)  { set_char(str, 250); } // small u, acute accent 
  else if CMP("&ucirc;", str)   { set_char(str, 251); } // small u, circumflex accent 
  else if CMP("&uuml;", str)    { set_char(str, 252); } // small u, dieresis or umlaut mark 
  else if CMP("&yacute;", str)  { set_char(str, 253); } // small y, acute accent 
  else if CMP("&thorn;", str)   { set_char(str, 254); } // small thorn, Icelandic 
  else if CMP("&yuml;", str)    { set_char(str, 255); } // small y, dieresis or umlaut mark 
  
  else { 
#if DEBUG
		printf("No Latin1 %s\n", str); 
#endif
		return(0); } // found no latin1 entity
  return(1); // found latin1 entity 
} // end latin1
